
# fourth-rna.py  # for python 2.7
# call this program on a command line as: python fourth-rna.py name-of-RNA-file name-of-LP-output-file min-distance-between-paired-nucleotides
#
# Python program to generate the inequalities for the ILP formulation for the RNA folding problem. This program only generates a P(i,j) variable
# if nucleotides i and j are complementary, and  farther apart than a minimum distance between
# paired nucleotides in the RNA string. Also, this ILP formulation has inequalities to determine how many stacked quartets (not stacked pairs) there are
# in the computed fold, and the number of stacked quartets is included in the objective function. Later, we should weight matched pairs
# and stacked quartets differently. The variable Q(p,q) gets set to 1 if and only if (p,q) first pair in a  stacked quartet. So, for the innermost stacked pair 
# (x,y) in a stack, the variable Q(x,y) gets set to 0.

# Setting the min-distance-between-paired-nucleotides to 0 allows adjacent nucleotides to pair, if they are complementary.
#
# We try to make the Python programming as simple as possible. As a consequence,
# it generates some unneeded ILP variables, and some unneeded inequalities, but these
# do no harm, and will be cleaned up by the Gurobi preprocessor. 

# collect the information from the user
#
import sys  # this imports the module called `sys', which is a collection of programs
            # that facilitate system calls

arg1 = sys.argv[1]    # assign the first argument  to the variable arg1
arg2 = sys.argv[2]    # assign the second argument to the variable arg2
                      # The values in arg1 and arg2 are strings.
minD = sys.argv[3]    # minD is the minimum number of nucleotides that must be between a matched pair (usually around 3 to 5).
# print (minD)
minD = int(minD)      # convert from a string to an integer
listP = ""            # listP will sum the P variables
listQ = ""            # listQ will sum the Q variables

INFILE = open(arg1, "r")  # open the file specified by the value of arg1, to read from the file.
OUT = open(arg2, "w")

for RNA in INFILE:
    RNA =    RNA.rstrip()
    print ("You input the sequence: ", RNA)
    print (RNA)

length = len(RNA)
print (length)
length_for_range = length + 1

# legal is a (two-dimensional) list whose individual elements are also lists. The elements in those lists are numbers, either 0 or 1.
# To reference an individual element in legal, say the j'th element in the i'th list, Python uses ``legal[i-1][j-1]". Remember that 
# indexing in Python begins with zero (Ugh!), so the i'th list of legal is actually list i-1 (UGH!); and the j'th element
#in list i-1 is actually element j-1 (double UGH!).
# The following statement alerts Python that ``legal" is a two-dimensional list, and initializes all of the entries in legal to have value 0.

legal = [[0 for x in range(1, length_for_range + 1)] for y in range(1, length_for_range + 1)] 
#print (legal)

# The following lines determine which nucleotide pairs are legal to pair.

for i in range(1, length_for_range - 1):  
  for j in range(i, length_for_range):
#           print (j-i, minD)
      if (j-i  > minD):  # only consider allowing i to pair with j is they are at least minD nucleotides between them.

#           print (i,j,RNA[i-1],RNA[j-1])
           if (RNA[i-1] == 'A') and (RNA[j-1] == 'U'):
                   legal[i][j] = 1                     # setting this value to 1 indicates that pair (i,j) in the RNA string are legal to pair.
#                   OUT.write ("P(%d,%d) = 0 \n" % (i,j))

           if (RNA[i-1] == 'U') and (RNA[j-1] == 'A'):
                   legal[i][j] = 1
#                   OUT.write ("P(%d,%d) = 0 \n" % (i,j))

           if (RNA[i-1] == 'C') and (RNA[j-1] == 'G'):
                   legal[i][j] = 1
#                   OUT.write ("P(%d,%d) = 0 \n" % (i,j))

           if (RNA[i-1] == 'G') and (RNA[j-1] == 'C'):
                   legal[i][j] = 1
#                   OUT.write ("P(%d,%d) =  0 \n" % (i,j))
           legal[j][i] = legal[i][j]

# The following generates the objective function

#print (legal)
OUT.write ("Maximize \n")

for i in range(1, length_for_range):   # accumulate a list of the P and Q variables for the legal pairs 
  for j in range(i, length_for_range):
#      print (i,j, legal[i][j])
      if (legal[i][j] == 1):
           listP = listP + "+ P(%d,%d) " % (i,j)
           listQ = listQ + "+ Q(%d,%d) " % (i,j)

# For exercise XXX in Chapter 6,
# choose the appropriate objective function from the following three choices. Comment out the two choices
# you don't want, and leave the choice you do want, uncommented, so it is used.

OUT.write ("P + Q \n") 
#OUT.write ("P \n")
#OUT.write ("Q \n")

OUT.write ("\nsuch that \n")
OUT.write (listP + "- P = 0 \n\n") # produce an equality that sets ILP variable P to the sum of the P(i,j) variables.
OUT.write (listQ + "- Q = 0 \n\n")  # produce an equality that sets ILP variable Q to the sum of the Q(i,j) variables.

# The following lines generate the inequalities to ensure
# that each nucleotide is paired to at most one other nucleotide

for i in range(1,length_for_range):
   inequality = ""

   for j in range(1, i):    # this was changed from i+1 Nov. 10, 2017 - check if bugs appear
        if (legal[j][i]):
            inequality = inequality + " + P(%d,%d)" % (j,i)

   for j in range(i+1, length_for_range):
        if (legal[i][j]):
            inequality = inequality + " + P(%d,%d)" % (i,j)

   inequality = inequality + ' <= 1'
   OUT.write (inequality)           		
   OUT.write ("\n\n")

# The following lines generate the inequalities to ensure
# that no pairs cross.

for h in range(1, length_for_range - 3):
    for i in range(h+1, length_for_range - 2):
        for j in range(i+1, length_for_range - 1):
            for k in range(j+1, length_for_range):
               if ((legal[h][j] == 1) and (legal[i][k])): 
                    OUT.write ("P(%d,%d) + P(%d,%d) <= 1 \n" % (h,j,i,k))

# The following lines generate the inequalities to add in possible stacking
# values.

for i in range (1, length_for_range): 
    for j in range(i + minD + 1,  length_for_range): 
          ip1 = i + 1;
          jm1 = j - 1;
          if (legal[i][j] == 1):
             if ((legal[ip1][jm1]) and (ip1 < jm1)): 
                                                                      # if Q(i,j) is set to 1, both of the variables P(i,j) and P(i+1,j-1)  variables
                                                                      # must also be set to 1. So Q(i,j) cannot
                                                                      # be 1 unless both  pairings occur.
        
                 OUT.write ("2 Q(%d,%d) - P(%d,%d) - P(%d,%d) <= 0 \n" % (i,j,i,j,ip1,jm1))

                                                                      # create the inequalities to enforce the converse, although it is
                                                                      # redundant. You can test this fact by commenting these out. Does it make
                                                                      # difference in the Gurobi execution times?

                 OUT.write ("P(%d,%d) + P(%d,%d) -  Q(%d,%d) <= 1 \n" % (i,j,ip1,jm1,i,j))
             else:
                 OUT.write ("Q(%d,%d) = 0 \n" % (i,j))    # The purpose of this is to say that if (i,j) is a legal pair, but (i+1,j-1) is not, then
                                                          # set Q(i,j) to 0. This compensates for the fact that Q(i,j) may be part of the objective function.

OUT.write("\n")
OUT.write ("binary \n")
for i in range(1, length_for_range):
   for j in range(i, length_for_range):
      if (legal[i][j]):
        OUT.write ("P(%d,%d) \n" % (i,j))
        OUT.write ("Q(%d,%d) \n" % (i,j))
OUT.write("\n")
OUT.write ("general\nP\nQ\n")
OUT.write  ("end \n")
print "The ILP is in file %s" % arg2
INFILE.close()
