# tspjulia.py
#
# Modified by Julia Matsieva <jmatsieva@ucdavis.edu>
# 6/7/16
#
# This program modifies the Gurobi program tsp.py, to read in
# a user-supplied edge-distance matrix. The original tsp.py only
# generates random Euclidian-distance problems.
# The ILP generated is only for undirected graphs, i.e., the cost for
# using an edge is the same in both directions. 
#
# If your argument is a distance matrix, run the program with -d. 
# Examples of how to run the program:

#    python tspjulia.py saharaData
#    python tspjulia.py -d outD
#    python tspjulia.py drilling280

#    python tspjulia.py -h will show help.

# Gurobi holds the copyright for the original code.
# See tsp.py for the original code.
# Gurobi code used by permission from Gurobi Optimization, Inc.
#
#!/usr/bin/python

# Copyright 2016, Gurobi Optimization, Inc.

# Solve a traveling salesman problem on a randomly generated set of
# points using lazy constraints.   The base MIP model only includes
# 'degree-2' constraints, requiring each node to have exactly
# two incident edges.  Solutions to this model may contain subtours -
# tours that don't visit every city.  The lazy constraint callback
# adds new constraints to cut them off.

import sys
import math
import random
from gurobipy import *


# Callback - use lazy constraints to eliminate sub-tours

def subtourelim(model, where): 
    if where == GRB.Callback.MIPSOL:
        selected = []
        # make a list of edges selected in the solution
        for i in range(n):
            sol = model.cbGetSolution([model._vars[i,j] for j in range(n)])
            selected += [(i,j) for j in range(n) if sol[j] > 0.5]
        # find the shortest cycle in the selected edge list
        tour = subtour(selected)
        if len(tour) < n:
            # add a subtour elimination constraint
            expr = 0
            for i in range(len(tour)):
                for j in range(i+1, len(tour)):
                    expr += model._vars[tour[i], tour[j]]
            model.cbLazy(expr <= len(tour)-1)
            print "Another Subtour Elimination \n"; 

# Euclidean distance between two points

def distance(points, i, j):
    global distanceMatrix
    if distanceMatrix:
        return distanceMatrix[i][j]

    dx = points[i][0] - points[j][0]
    dy = points[i][1] - points[j][1]
    return math.sqrt(dx*dx + dy*dy)


# Given a list of edges, finds the shortest subtour

def subtour(edges):
    visited = [False]*n
    cycles = []
    lengths = []
    selected = [[] for i in range(n)]
    for x,y in edges:
        selected[x].append(y)
    while True:
        current = visited.index(False)
        thiscycle = [current]
        while True:
            visited[current] = True
            neighbors = [x for x in selected[current] if not visited[x]]
            if len(neighbors) == 0:
                break
            current = neighbors[0]
            thiscycle.append(current)
        cycles.append(thiscycle)
        lengths.append(len(thiscycle))
        if sum(lengths) == n:
            break
    return cycles[lengths.index(min(lengths))]


# Parse argument

from argparse import ArgumentParser

parser = ArgumentParser()

parser.add_argument("arg", action="store", help="Use integer to generate random data or a filename.")

parser.add_argument("-d", action="store_true", default=False, dest="distance", 
        help="indicates that file is a distance matrix")

parser.add_argument("-v", action="store_true", default=False, dest="verbose",
        help="prints a ton of stuff")

args = parser.parse_args()

# need to set these in all the options
points = []
n = 0

distanceMatrix = None

if args.distance:
#    print "File is distance matrix."
    distanceMatrix = []
    f = open(args.arg)
    for line in f:
        print "%s" % line #DG
        l = map(float, line[:-2].split(" "))
        distanceMatrix.append(l)
        n += 1
    f.close()
    points = range(n)
else:
    try:
        n = int(args.arg)
        random.seed(1)
        print "Generating %s random points." % n
        for i in range(n):
            points.append((random.randint(0,100),random.randint(0,100)))

    except ValueError:
        print "Reading point file %s." % args.arg
        f = open(args.arg)
        for line in f:
            try:
                L = filter(lambda x: x!="", line.split(" "))
                x, y = map(float, L[1:3]) 
                points.append((x,y))
                n+=1
            except ValueError:
                print "Ignoring line:", line,
                continue
        print
        f.close() 

# Create n random points
print "points:", len(points)
print

m = Model()

# Create variables

vars = {}
for i in range(n):
    for j in range(i+1):
        vars[i,j] = m.addVar(obj=distance(points, i, j), vtype=GRB.BINARY,
                             name='e'+str(i)+'_'+str(j))
        vars[j,i] = vars[i,j]
m.update()

# Add degree-2 constraint, and forbid loops

for i in range(n):
    m.addConstr(quicksum(vars[i,j] for j in range(n)) == 2)
    vars[i,i].ub = 0
m.update()


# Optimize model

m._vars = vars
m.params.LazyConstraints = 1
m.optimize(subtourelim)

solution = m.getAttr('x', vars)
selected = [(i,j) for i in range(n) for j in range(n) if solution[i,j] > 0.5]
assert len(subtour(selected)) == n


if args.verbose:
    w = len(str(m.objVal)) 
    nn = len(str(n))+1

    total = 0

    for i in range(n-1):
        x, y = subtour(selected)[i:i+2]
        d = round(distance(points, x, y), 4)
        total += d
        print str(i).rjust(nn), str((x+1, y+1)).rjust(nn*2+4),
        print str(d).rjust(w), str(total).rjust(w)

    x = subtour(selected)[n-1]
    y = subtour(selected)[0]
    d = round(distance(points, x, y))
    total += d
    print str(i+1).rjust(nn), str((x+1, y+1)).rjust(nn*2+4), 
    print str(d).rjust(w), str(total).rjust(w)

print('')
print('Optimal tour: %s' % str(map(lambda x: x+1, subtour(selected))))
print('Optimal cost: %g' % m.objVal)
print('')
